////////////////////////////////////////////////
//          REAL TIME FUR RENDERING
// Program running a shader to create real time fur
// Only works on Nvidia card
// By Jonas Nilsson
// jonni957@student.liu.se
//
// The Perlin noise and the sphere algoritm
//  used is made by Stefan Gustavsson
////////////////////////////////////////////////


#include <GL/glew.h>
#include <GL/glut.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "resource/glslnoisetext.h"
#include "resource/textfile.h"

//Fps counter and other stuff to keep track of time
//and performance
float t0 = 0.0;
float FPSmin=10000;
float FPSmax=0;
float fps=0;
int frames = 0;
int totframes =0;
char titlestring[200];
GLfloat time = 0.0;

//Initial values used for the fur
int layers = 30;
float furLength = 0.5;
float layerdistance = furLength/layers;
float freq = 100.0;
float blur = 1;

//Mouse variables
int mousex,mousey;
bool leftmouse,middlemouse;

//Inital values used for rotation and translation
static float zoom = 4;
static float rot = 60;
static float rot2 = 0;
static float transy = 0;
static float transx = 0;
static GLfloat size = 1;

//Flags 
static int object = 1;
static int loop = 0;
static int timeloop = 1;
static int lightrot= 1;
static int rect = 1;
static int wind = 1;

//Menus
static int submenu;
static int mainmenu;

//Shader files to start with
char* vertexShaderFile = "windshader.vert";
char* fragmentShaderFile = "noise1.frag";

/* GLUT callback Handlers */
GLfloat lightDir[3] = {1,0,0};

//Lighting colors
const GLfloat light_ambient[]  = { 0.2f, 0.2f, 0.2f, 1.0f };
const GLfloat light_diffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
const GLfloat light_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
const GLfloat light_position[] = { 1.0f, 2.0f, 1.0f, 1.0f };
//Material colors
const GLfloat mat_ambient[]    = { 0.6f, 0.5f, 0.4f, 1.0f };
const GLfloat mat_diffuse[]    = { 0.7f, 0.6f, 0.5f, 1.0f };
const GLfloat mat_specular[]   = { 1.0f, 1.0f, 1.0f, 1.0f };
const GLfloat high_shininess[] = { 1000.0f };

//Textures
GLuint noiseTextureID;
GLuint filterTexture;

//Shader programs
GLuint main_program;
GLuint rectangle_shader;
GLuint filter_shader;

//Display list
GLuint thespheres;

//Error handling variables
GLenum errcode;
char errstring[2000];

//Functions
void setShaderProgram(GLuint * program, char * shadername);
void key(unsigned char key, int x, int y);
void menu(int value);
void createMenu(void);
static void resize(int width, int height);
void showFPS();
static void display(void);
static void idle(void);


////////////////////////////////////////////////
//               SHADERS
//
////////////////////////////////////////////////
//Loader shader program
//Report errrors
void setShaderProgram(GLuint * program, char * vertname,char * fragname)
{
char *vs,*fs;
*program = glCreateProgram();
GLint shaderstatus;
//Loading and compiling vertex shader	
          char path[100];
          strcpy(path,"shaders/");
          strcat(path,vertname);
		GLuint vertex_shader = glCreateShader(GL_VERTEX_SHADER);
		
	    vs = textFileRead(path);
		const char * vv = vs;
  		glShaderSource(vertex_shader, 1, &vv,NULL);
	    free(vs);
       	glCompileShader(vertex_shader);
		glAttachShader(*program,vertex_shader);
       glGetShaderiv(vertex_shader,GL_COMPILE_STATUS,&shaderstatus);
       if(shaderstatus != GL_TRUE)
       {
         glGetShaderInfoLog(vertex_shader,sizeof(errstring),NULL,errstring);
         MessageBox (NULL, (char*)errstring,"Vertex shader error", 0);
       }    
//Loading and compiling fragment shader
        GLuint fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
        strcpy(path,"shaders/");	
        strcat(path,fragname);
		fs = textFileRead(path);
		const char * ff = fs;
		glShaderSource(fragment_shader, 1, &ff,NULL);
        free(fs);
		glCompileShader(fragment_shader);
    	glAttachShader(*program,fragment_shader);
       glGetShaderiv(fragment_shader,GL_COMPILE_STATUS,&shaderstatus);
       if(shaderstatus != GL_TRUE)
       {
         glGetShaderInfoLog(fragment_shader,sizeof(errstring),NULL,errstring);
         MessageBox (NULL, (char*)errstring,"Fragment shader error", 0);
       }
		glLinkProgram(*program);
        glUseProgram(*program);
        
}

////////////////////////////////////////////////
//               DISPLAY
//
////////////////////////////////////////////////

//Resize function
//The setup for camera and viewport
static void 
resize(int width, int height)
{
    const float ar = (float) width / (float) height;
    
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
        // Set viewport. This is the pixel rectangle we want to draw into.
    glViewport( 0, 0, width, height ); // The entire window

    // Select and setup the projection matrix.
    glMatrixMode(GL_PROJECTION); // "We want to edit the projection matrix"
    glLoadIdentity(); // Reset the matrix to identity
    // 45 degrees FOV, same aspect ratio as viewport, depth range 0.1 to 1000
    gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f );

    // Select and setup the modelview matrix.
    glMatrixMode( GL_MODELVIEW ); // "We want to edit the modelview matrix"
    glLoadIdentity(); // Reset the matrix to identity
    // Look from 0,-4,0 towards 0,0,0 with Z as "up" in the image
    gluLookAt( 0.0f, -4.0f, 0.0f,  // Eye position
               0.0f, 0.0f, 0.0f,   // View point
               0.0f, 0.0f, 1.0f ); // Up vector
}

//Keeps track of fps and display information in the titlebar
void showFPS() {
  // If one second has passed, or if this is the very first frame
    if( (time-t0) > 1.0 || totframes ==0)
    {
        fps = (float)frames / (time-t0);
        if(totframes>10)
        if(fps<FPSmin) FPSmin=fps;
        if(fps>FPSmax) FPSmax=fps;
        
        t0 = time;

        frames = 0;

    }
    sprintf(titlestring, "Fur rendering  Layers:%d,Length:%.1f,Freq:%.0f,Blur:%.1f,  '%s' '%s'  %.1f FPS",layers,furLength,freq,blur,vertexShaderFile,fragmentShaderFile,fps);
    glutSetWindowTitle(titlestring); 
    
    frames ++;
    totframes+=frames;
}


/*
 * drawTexturedSphere(r, segs) - Draw a sphere centered on the local
 * origin, with radius "r" and approximated by "segs" polygon segments,
 * with texture coordinates in a latitude-longitude mapping.
 *///by Stefan Gustavsson
void drawTexturedSphere(float r, int segs) {
  int i, j;
  float x, y, z, z1, z2, R, R1, R2;

  // Top cap
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0,0,1);
  glTexCoord2f(0.5f,1.0f); // This is an ugly (u,v)-mapping singularity
  glVertex3f(0,0,r);
  z = cos(M_PI/segs);
  R = sin(M_PI/segs);
    for(i = 0; i <= 2*segs; i++) {
      x = R*cos(i*2.0*M_PI/(2*segs));
      y = R*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(x, y, z);
      glTexCoord2f((float)i/(2*segs), 1.0f-1.0f/segs);
      glVertex3f(r*x, r*y, r*z);
    }
  glEnd();  

  // Height segments
  for(j = 1; j < segs-1; j++) {
    z1 = cos(j*M_PI/segs);
    R1 = sin(j*M_PI/segs);
    z2 = cos((j+1)*M_PI/segs);
    R2 = sin((j+1)*M_PI/segs);
    glBegin(GL_TRIANGLE_STRIP);
    for(i = 0; i <= 2*segs; i++) {
      x = R1*cos(i*2.0*M_PI/(2*segs));
      y = R1*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(x, y, z1);
      glTexCoord2f((float)i/(2*segs), 1.0f-(float)j/segs);
      glVertex3f(r*x, r*y, r*z1);
      x = R2*cos(i*2.0*M_PI/(2*segs));
      y = R2*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(x, y, z2);
      glTexCoord2f((float)i/(2*segs), 1.0f-(float)(j+1)/segs);
      glVertex3f(r*x, r*y, r*z2);
    }
    glEnd();
  }

  // Bottom cap
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0,0,-1);
  glTexCoord2f(0.5f, 1.0f); // This is an ugly (u,v)-mapping singularity
  glVertex3f(0,0,-r);
  z = -cos(M_PI/segs);
  R = sin(M_PI/segs);
    for(i = 2*segs; i >= 0; i--) {
      x = R*cos(i*2.0*M_PI/(2*segs));
      y = R*sin(i*2.0*M_PI/(2*segs));
      glNormal3f(x, y, z);
      glTexCoord2f(1.0f-(float)i/(2*segs), 1.0f/segs);
      glVertex3f(r*x, r*y, r*z);
    }
  glEnd();
}

//The sphere list is initiated with calls to draw all the spheres 
void initSpheresList(GLuint *listID, GLdouble scale)
{
  
  glNewList(*listID, GL_COMPILE);
      for(int i = 0; i<layers ; i++)
              {
              drawTexturedSphere(1+(layerdistance*i),50);
              glUniform1i(glGetUniformLocation(main_program,"level"),i);
              }
  glEndList();
}


//Function to draw a rectangle with texture coordinates
void drawTexturedRectangle(float x1,float y1,float x2,float y2)
{
  glBegin(GL_POLYGON); 
  glMultiTexCoord2f(GL_TEXTURE0,0.0,0.0); glVertex2f(x1, y1);
  glMultiTexCoord2f(GL_TEXTURE0,1.0,0.0); glVertex2f(x2, y1);
  glMultiTexCoord2f(GL_TEXTURE0,1.0,1.0); glVertex2f(x2, y2);
  glMultiTexCoord2f(GL_TEXTURE0,0.0,1.0); glVertex2f(x1, y2);
  glEnd();
}

//Function called when bluring is on
//The framebuffer is copied to a texture and rendered to
//a rectangle with a gauss blur shader. 
void blurTexture()
{  
     //The texture to store the buffer in is binded
     glBindTexture(GL_TEXTURE_2D,filterTexture);
     //The entire buffer is copied to texture
     glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0,  glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT),0);
     
     //The filter program is activated and the current texture is sent
     glUseProgram(filter_shader);
     glUniform1i(glGetUniformLocation(filter_shader,"texture"),0);    
     //By send a smaller width and height makes it sample with smaller or larger step sizes
     //Smaller step sizes will result in lessamount of blur
     glUniform1i(glGetUniformLocation(filter_shader,"w"),(int)(glutGet(GLUT_WINDOW_WIDTH)/blur));
     glUniform1i(glGetUniformLocation(filter_shader,"h"),(int)(glutGet(GLUT_WINDOW_HEIGHT)/blur));

     //We don't want any lighting or blending affect to rendered rectangle
     glDisable(GL_LIGHTING);
     glDisable(GL_BLEND);

     //The view is changed to easy be able to draw a quad exactly covering
     //the whole view
     glLoadIdentity();
     glMatrixMode(GL_PROJECTION);
     glPushMatrix();
     glLoadIdentity();
     gluOrtho2D(0.0,glutGet(GLUT_WINDOW_WIDTH),0.0,glutGet(GLUT_WINDOW_HEIGHT));

     //Quad is drawn covering the whole view
     glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     drawTexturedRectangle(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
     
     //Program is deactivated
     glUseProgram(0);

     //Everything is restored     
     glPopMatrix();
     glMatrixMode(GL_MODELVIEW);
     glEnable(GL_LIGHTING);
     glEnable(GL_BLEND);
}

//The display function to run in each frame
static void 
display(void)
{   
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glEnable(GL_TEXTURE_2D);
   
   time = glutGet(GLUT_ELAPSED_TIME) / 1000.0;    
  
   glLoadIdentity();
 
   //If light rotation is triggered will the light be rotated according to time
   if(lightrot){
              glPushMatrix(); // Transforms to animate the light:
              glRotatef(30.0f*time, 0.0f, 1.0f, 0.0f);
              glTranslatef(3.0f, 1.0f, 0.0f); // Orbit around Z, 5 units from origin
              float lightpos1[4]={0.0f, 0.0f, 0.0f, 1.0f}; // Origin, in hom. coords
              glLightfv(GL_LIGHT0, GL_POSITION, lightpos1); // Set light position
              glPopMatrix(); // Revert to initial transform
              }
    //Else the light has a fix position
    else glLightfv(GL_LIGHT0, GL_POSITION, light_position); 
    
    //The scene is translated
    glTranslatef(transy*0.1,transx*0.1,-zoom);     
    
    //If the background rectangle is top be drawn
    //The objects rotations do not affect the rectangles position
    if(rect)
    {
             glUseProgram(rectangle_shader);
             glUniform1f(glGetUniformLocation(rectangle_shader,"frequence"),freq);
             glUniform1f(glGetUniformLocation(rectangle_shader,"distortion"),(rot+rot2)/2);
             glUniform1i(glGetUniformLocation(rectangle_shader,"layers"),layers);
             drawTexturedRectangle(1+furLength,1+furLength,-(1+furLength),-(1+furLength));
             glUseProgram(0);
    }
    
    //The object is rotated
    glRotatef(rot, 0.0f, 1.0f, 0.0f );
    glRotatef(rot2, 1.0f, 0.0f, 0.0f );   
    
    //The objects texture and shader is activated
    glBindTexture(GL_TEXTURE_2D, noiseTextureID);
    glUseProgram(main_program);
    glUniform1i(glGetUniformLocation(main_program,"noiseTexture"),0);
    glUniform1i(glGetUniformLocation(main_program,"layers"),layers);
    glUniform1f(glGetUniformLocation(main_program,"freq"),freq);
    glUniform1f(glGetUniformLocation(main_program,"shelldist"),layerdistance);
    glUniform1f(glGetUniformLocation(main_program,"time"),10*time);
    //The object is rendered depending on which is choosen
    if(object)
    {               
                    glCallList(thespheres);
    }    
    else           
    {
                   for(int i = 0; i<layers; i++)
                         {
                        glutSolidTeapot(1+(float)i*layerdistance);
                        glUniform1i(glGetUniformLocation(main_program,"level"),i);
                        }
    }
    //The program is deactivated
    glUseProgram(0);
    
    //Experiment with bluring
    //If bluring is on will the blur function be called
    if(blur!=0) blurTexture();
    
    //Display loop is done
    glFlush();
    glutSwapBuffers();
}

//Idle function
static void 
idle(void)
{
 showFPS();
 glutPostRedisplay();
}


////////////////////////////////////////////////
//                  INERACTIONS
//
////////////////////////////////////////////////


//Menu option function
//Each menu item corresponds to an integer.
void menu(int value)
{
 switch(value)
 {     
    case 0:
         exit(0);
         break;
    
    case 1:
         if (lightrot==1) lightrot=0;
         else lightrot=1;
         break;
    
    case 2:
         if (object==1) object=0;
         else object=1;
         break;
    
    case 3:
        setShaderProgram(&main_program,vertexShaderFile,fragmentShaderFile);
        setShaderProgram(&filter_shader,"minimal.vert","filter.frag");
        setShaderProgram(&rectangle_shader,"standard.vert","rectangle.frag");
        glUniform1i(glGetUniformLocation(main_program,"noiseTexture"),noiseTextureID);
        glUniform1i(glGetUniformLocation(main_program,"layers"),layers);
        glUniform1i(glGetUniformLocation(rectangle_shader,"noiseTexture"),noiseTextureID);
        glUniform1i(glGetUniformLocation(rectangle_shader,"layers"),layers);
        glUseProgram(main_program);
        break;
        
    case 4:
        if(rect==1) rect = 0;
        else rect = 1;
        break;
    
    case 5:
        if(wind!=0)
        { 
        vertexShaderFile="standard.vert";
        menu(3);
        wind = 0;
        }
        else{
        wind = 1;
        vertexShaderFile="windshader.vert";
        menu(3);
        }     
        break;
    
    case 6:
        if(blur!=0) blur = 0;
        else blur = 1;
        break;  
 }
}
///Menu to choose shader function
void shaderMenuFunc(int value)
{
   switch(value)
   {
                case 1:
                     fragmentShaderFile="base.frag";
                     menu(3);
                     break;
                case 2:
                     fragmentShaderFile="antialias.frag";
                     menu(3);
                     break;
                case 3:
                     fragmentShaderFile="noise1.frag";
                     menu(3);
                     break;
                case 4:
                     fragmentShaderFile="noise2.frag";
                     menu(3);
                     break;
   }    

}

//Menu creation
void createMenu(void)
{
  submenu = glutCreateMenu(shaderMenuFunc);
  
  glutAddMenuEntry("Base shader 'base.frag'",1);
  glutAddMenuEntry("Distortion noise 'noise1.frag'",3);

  glutAddMenuEntry("Frequency noise 'noise2.frag'",4);    
  glutAddMenuEntry("Antialiased shader (doesn't work yet) 'antialias.frag'",2);    
  
  mainmenu = glutCreateMenu(menu);
  glutAddSubMenu("Choose shader",submenu);
  glutAddMenuEntry("Toggle Wind",5);
  glutAddMenuEntry("Toggle Light rotation",1);
  glutAddMenuEntry("Toggle Blur", 6);
  glutAddMenuEntry("Toggle background",4);
  glutAddMenuEntry("Change Object", 2);
  glutAddMenuEntry("Reload shader",3);
  
  glutAddMenuEntry("Quit", 0);
  
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}


//Mouse interaction function.
//Runs when a mouse button is clicked
void
mouse(int button, int state, int x, int y)
{

     mousex=x;
     mousey=y;
     
     leftmouse=(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN);

     middlemouse=(button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN);       
}
//Mouse motion function
//Runs when the mouse is moved over the window.
//Makes the bocjet rotate when the left button is held
void processMouseActiveMotion(int x, int y)
{
     if(leftmouse)
     {
     rot+=(((float)x-(float)mousex)*0.5);                
     rot2+=(((float)mousey-(float)y)*0.5);    
     }
     else if(middlemouse)
     {
     zoom-=(((float)mousey-(float)y)*0.1);    
     }
     mousex=x;
     mousey=y;
     glutPostRedisplay();
}


//Keyhandler. Handles input from the keyboard.
void key(unsigned char key, int x, int y)
{
    switch (key) 
    {
    //Esc exits program
        case 27 : 
             exit(0);
             break;
             
    //+ and - increase and decrease to number of layers used.
    //A new displaylist have to be generated                
        case '+':
            layers++;
            layerdistance = furLength/layers;
            initSpheresList(&thespheres, 1.0);
            break;
        case '-':
            layers--;
            layerdistance = furLength/layers;
            initSpheresList(&thespheres, 1.0);
            break;

    //* and / increase and decrease to distance between layers.
    //A new displaylist have to be generated                            
        case '*':
            furLength+=0.1;
            layerdistance = furLength/layers;
            initSpheresList(&thespheres, 1.0);
            break;
        case '/':
            if(furLength-0.1 > 0)
             furLength-=0.1;
            layerdistance = furLength/layers;
            initSpheresList(&thespheres, 1.0);
            break;
    
    //* and / increase and decrease the frequency used by the shader.
        case '9':
            freq += 10;
            initNoiseTexture(&noiseTextureID,(int)(64/(100/freq)));
            break;
        case '3':
            freq -= 10;
            initNoiseTexture(&noiseTextureID,(int)(64/(100/freq)));
            break;

    //7 and 1 increase and decrease the amount of blur.
        case '7':
            blur += 0.1;
            break;
        case '1':
            if(blur>0.1)
               blur -= 0.1;
            else blur = 0.0;
            break;
    }

    glutPostRedisplay();
}




/* Program entry point */
int 
main(int argc, char *argv[])
{
 //Glut initialisations
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(640,480);
    glutInitWindowPosition(10,10);
    glutCreateWindow("");
    glutReshapeFunc(resize);
    glutDisplayFunc(display);
    glutKeyboardFunc(key);
    glutIdleFunc(idle);
    glutMouseFunc(mouse);
    glutMotionFunc(processMouseActiveMotion);
    glClearColor(1,1,1,1);
   
   //Enable what we want to use
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_LIGHT0);
    glEnable(GL_NORMALIZE);
    glEnable(GL_LIGHTING);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    
   //We only got one light
    glLightfv(GL_LIGHT0, GL_AMBIENT,  light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
   //And one object with material
    glMaterialfv(GL_FRONT, GL_AMBIENT,   mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE,   mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR,  mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);

    //Check with glew if the shaders can be run. 
    //If not exit program
    glewInit();
   	if (!glewIsSupported("GL_VERSION_2_0"))
   	{
		MessageBox (NULL, "No Opengl 2.0 support, shaders can not be used" , "Error", 0);
	}
	
	//Load shaders
    setShaderProgram(&main_program,vertexShaderFile,fragmentShaderFile);
    setShaderProgram(&filter_shader,"minimal.vert","filter.frag");
    setShaderProgram(&rectangle_shader,"standard.vert","rectangle.frag");
    
    glUniform1i(glGetUniformLocation(main_program,"noiseTexture"),noiseTextureID);
    glUniform1i(glGetUniformLocation(main_program,"layers"),layers);
    glUniform1i(glGetUniformLocation(rectangle_shader,"noiseTexture"),noiseTextureID);
    glUniform1i(glGetUniformLocation(rectangle_shader,"layers"),layers);

    //Generate displaylist for the spheres
    thespheres = glGenLists(1);
    initSpheresList(&thespheres, 1.0);
    
    //Genrerate a perlin noise texture
    initNoiseTexture(&noiseTextureID,64);
    
    //Generate a texture to be used when copying framebuffer
    glGenTextures(1,&filterTexture);
    glBindTexture(GL_TEXTURE_2D,filterTexture);
    glTexImage2D(GL_TEXTURE_2D,0 ,3 ,glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT), 0 , GL_RGB, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 
    //Create menu
    createMenu();
    
    //and lets go
    glutMainLoop();

    return EXIT_SUCCESS;
}

