/////////////////////////////////////
// 	 Rectangle fragment shader
//  Mimic behavior of the fur shader 
// 		on the sphere      
//  This shader needs more work... 
// or a more likely a total rewrite
//
//           by Jonas Nilsson
//       jonni957@student.liu.se
/////////////////////////////////////


uniform int layers; //Total number of layers
uniform float distortion; //Aproximation of the rotation of the object
uniform float frequence; //The frequency of the fur

varying vec3 lightDir,normal,tangent,binorm,halfv; //Used in lighting calculations
varying vec2 v_texCoord2D; //Texture coordinat

void main (void)
{
	
	//Ambient and diffuse color is calcuated from the material
	vec4 amb = gl_FrontMaterial.ambient*gl_LightSource[0].ambient;
	vec4 dif = gl_FrontMaterial.diffuse*gl_LightSource[0].diffuse;
    vec4 globalAmbient = gl_LightModel.ambient * gl_FrontMaterial.ambient;
	
	//The frequency used is half the frequency in the sphere
	float freq =frequence/2;

	//Calculate the distance to the center of the polygon
	float l = length(vec2(v_texCoord2D.s-0.5,v_texCoord2D.t-0.5));
	
	//Calculate the dot produkt af the texture coordiante and a vector in the circle around the centre
	float d = dot(normalize(vec2(v_texCoord2D.s-0.5,v_texCoord2D.t-0.5)),vec2(0,1));
	
	//Calculate the angle for the coordinate in a cirkel around the center
	float angle = acos(d);
   	
	//Calculate a step value to be used in the alpha calculation
	//this depends on a sin function of the angle and frequency
	// a higher value than 0.45 results in a smaller circle, a bigger results in a bigger circle.
	//This function could most probably be calculated more easly and with a better result using some other method
	float radius = 0.45 / clamp(abs(sin((angle+distortion)*freq)),0.92,0.96);
	
	//The alpha value is calculated from a smoothstep of the radius depending on the length 
	float alpha = smoothstep(radius,radius+0.11,(1.0-l)); 
	
	//As this is one polygon must the normals be perbutrated.
	//The amount to perputrate the normal depends on the the dot product. 
	vec3 U = dot(normalize(vec2(v_texCoord2D.s-0.5,v_texCoord2D.t-0.5)),vec2(1,0)) * cross(binorm,normal); 
	vec3 V = d * cross(tangent,normal);
	
	//The new normal is calculated by just adding the U and V vektor to the normal. And normalize the result
	vec3 nynorm = normalize(normal+U+V);
		
	//A new ambient color is calculated depending on the layer. 
	//This is a simple little effect to amplify the feeling of depth in the fur.
	vec4 color = (amb + globalAmbient);

	//Calculating dot product with the new normal. The angle of incomming light
	float NdotL = dot(-lightDir,nynorm);
	//If NdotL > 0.0 will incomming light be reflected, 
    //resulting in specular and diffuse
	if (NdotL > 0.0) 
    {
        color+= dif*NdotL;
      	vec3 halfvec = normalize(halfv);
		// normalize halfv and then compute the 
		// cosine (dot product) with the newv normal 
		//Halfvec = vector between lightdirection and camera
		//NdotV angle between halfvector and normal, resulting in specular highlight
		float NdotHV = max(dot(nynorm, halfvec),0.0);
		color += gl_FrontMaterial.specular * 
					gl_LightSource[0].specular * 
					pow(NdotHV, gl_FrontMaterial.shininess);
	}
	
	//Set the w component of the color to the alpha value
    color.w = alpha;
	
	//And were'r done
	gl_FragColor = color;   	

}	
